@@ -17,7 +17,7 @@
  * with this program; if not, write to the Free Software Foundation, Inc.,
  * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  *
- * (C) Copyright 2008 - 2010 Red Hat, Inc.
+ * (C) Copyright 2008 - 2011 Red Hat, Inc.
  */
 
 #include <string.h>
@@ -26,6 +26,7 @@
 #include <arpa/inet.h>
 
 #include <gtk/gtk.h>
+#include <gdk/gdkkeysyms.h>
 #include <glib/gi18n.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
@@ -86,7 +87,12 @@ typedef struct {
 	gboolean window_added;
 
 	/* Cached tree view entry for editing-canceled */
-	char *last_edited;
+	/* Used also for saving old value when switching between cells via mouse
+	 * clicks - GTK3 produces neither editing-canceled nor editing-done for
+	 * that :( */
+	char *last_edited; /* cell text */
+	char *last_path;   /* row in treeview */
+	int last_column;   /* column in treeview */
 } CEPageIP6Private;
 
 #define METHOD_COL_NAME 0
@@ -529,6 +535,10 @@ cell_editing_canceled (GtkCellRenderer *renderer, gpointer user_data)
 
 		ce_page_changed (CE_PAGE (self));
 	}
+
+	g_free (priv->last_path);
+	priv->last_path = NULL;
+	priv->last_column = -1;
 }
 
 static void
@@ -545,8 +555,12 @@ cell_edited (GtkCellRendererText *cell,
 	guint32 column;
 	GtkTreeViewColumn *next_col;
 
+	/* Free auxiliary stuff */
 	g_free (priv->last_edited);
 	priv->last_edited = NULL;
+	g_free (priv->last_path);
+	priv->last_path = NULL;
+	priv->last_column = -1;
 
 	column = GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (cell), "column"));
 	gtk_tree_model_get_iter (GTK_TREE_MODEL (store), &iter, path);
@@ -556,7 +570,6 @@ cell_edited (GtkCellRendererText *cell,
 	column = (column >= COL_LAST) ? 0 : column + 1;
 	next_col = gtk_tree_view_get_column (priv->addr_list, column);
 	gtk_tree_view_set_cursor_on_cell (priv->addr_list, path, next_col, priv->addr_cells[column], TRUE);
-	gtk_widget_grab_focus (GTK_WIDGET (priv->addr_list));
 
 	gtk_tree_path_free (path);
 	ce_page_changed (CE_PAGE (self));
@@ -606,6 +619,85 @@ ip_address_filter_cb (GtkEntry *   entry,
 	g_free (result);
 }
 
+static void
+delete_text_cb (GtkEditable *editable,
+                    gint start_pos,
+                    gint end_pos,
+                    gpointer user_data)
+{
+	CEPageIP6 *self = CE_PAGE_IP6 (user_data);
+	CEPageIP6Private *priv = CE_PAGE_IP6_GET_PRIVATE (self);
+
+	/* Keep last_edited up-to-date */
+	g_free (priv->last_edited);
+	priv->last_edited = g_strdup (gtk_editable_get_chars (editable, 0, -1));
+}
+
+static gboolean
+cell_changed_cb (GtkEditable *editable,
+                 gpointer user_data)
+{
+	char *cell_text;
+	guint column;
+#if GTK_CHECK_VERSION(3,0,0)
+	GdkRGBA rgba;
+#else
+	GdkColor color;
+#endif
+	gboolean value_valid = FALSE;
+	const char *colorname = NULL;
+
+	cell_text = gtk_editable_get_chars (editable, 0, -1);
+
+	/* The Prefix column is 0..128 */
+	column = GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (user_data), "column"));
+	if (column == COL_PREFIX) {
+		guint32 prefix;
+		char *end;
+
+		prefix = strtoul (cell_text, &end, 10);
+		if (!end || *end || prefix == 0 || prefix > 128)
+			value_valid = FALSE;
+		else
+			value_valid = TRUE;
+	} else {
+		struct in6_addr tmp_addr;
+
+		if (inet_pton (AF_INET6, cell_text, &tmp_addr))
+			value_valid = TRUE;
+	}
+
+	/* Change cell's background color while editing */
+	colorname = value_valid ? "lightgreen" : "red";
+
+#if GTK_CHECK_VERSION(3,0,0)
+	gdk_rgba_parse (&rgba, colorname);
+	gtk_widget_override_background_color (GTK_WIDGET (editable), GTK_STATE_NORMAL, &rgba);
+#else
+	gdk_color_parse (colorname, &color);
+	gtk_widget_modify_base (GTK_WIDGET (editable), GTK_STATE_NORMAL, &color);
+#endif
+
+	g_free (cell_text);
+	return FALSE;
+}
+
+static gboolean
+key_pressed_cb (GtkWidget *widget,
+                GdkEvent *event,
+                gpointer user_data)
+{
+#if !GDK_KEY_Tab
+	#define GDK_KEY_Tab GDK_Tab
+#endif
+
+	/* Tab should behave the same way as Enter (finish editing) */
+	if (event->type == GDK_KEY_PRESS && event->key.keyval == GDK_KEY_Tab)
+		gtk_cell_editable_editing_done (GTK_CELL_EDITABLE (widget));
+
+	return FALSE;
+}
+
 static void
 cell_editing_started (GtkCellRenderer *cell,
                       GtkCellEditable *editable,
@@ -621,8 +713,12 @@ cell_editing_started (GtkCellRenderer *cell,
 		return;
 	}
 
+	/* Initialize last_path and last_column, last_edited is initialized when the cell is edited */
 	g_free (priv->last_edited);
 	priv->last_edited = NULL;
+	g_free (priv->last_path);
+	priv->last_path = g_strdup (path);
+	priv->last_column = GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (cell), "column"));
 
 	/* Need to pass column # to the editable's insert-text function */	
 	column = GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (cell), "column"));
@@ -632,6 +728,20 @@ cell_editing_started (GtkCellRenderer *cell,
 	g_signal_connect (G_OBJECT (editable), "insert-text",
 	                  (GCallback) ip_address_filter_cb,
 	                  user_data);
+
+	g_signal_connect_after (G_OBJECT (editable), "delete-text",
+	                        (GCallback) delete_text_cb,
+	                        user_data);
+
+	/* Set up handler for value verifying and changing cell background */
+	g_signal_connect (G_OBJECT (editable), "changed",
+	                  (GCallback) cell_changed_cb,
+	                  cell);
+
+	/* Set up key pressed handler - need to handle Tab key */
+	g_signal_connect (G_OBJECT (editable), "key-press-event",
+	                  (GCallback) key_pressed_cb,
+	                  user_data);
 }
 
 static void
@@ -694,6 +804,41 @@ routes_button_clicked_cb (GtkWidget *button, gpointer user_data)
 	gtk_widget_show_all (dialog);
 }
 
+static gboolean
+tree_view_button_pressed_cb (GtkWidget *widget,
+                             GdkEvent *event,
+                             gpointer user_data)
+{
+	CEPageIP6 *self = CE_PAGE_IP6 (user_data);
+	CEPageIP6Private *priv = CE_PAGE_IP6_GET_PRIVATE (self);
+
+	/* last_edited can be set e.g. when we get here by clicking an cell while
+	 * editing another cell. GTK3 issue neither editing-canceled nor editing-done
+	 * for cell renderer. Thus the previous cell value isn't saved. Store it now. */
+	if (priv->last_edited && priv->last_path) {
+		GtkTreeIter iter;
+		GtkListStore *store = GTK_LIST_STORE (gtk_tree_view_get_model (priv->addr_list));
+		GtkTreePath *last_treepath = gtk_tree_path_new_from_string (priv->last_path);
+
+		gtk_tree_model_get_iter (GTK_TREE_MODEL (store), &iter, last_treepath);
+		gtk_list_store_set (store, &iter, priv->last_column, priv->last_edited, -1);
+		gtk_tree_path_free (last_treepath);
+
+		g_free (priv->last_edited);
+		priv->last_edited = NULL;
+		g_free (priv->last_path);
+		priv->last_path = NULL;
+		priv->last_column = -1;
+	}
+
+	/* Ignore double clicks events. (They are issued after the single clicks, see GdkEventButton) */
+	if (event->type == GDK_2BUTTON_PRESS)
+		return TRUE;
+
+	gtk_widget_grab_focus (GTK_WIDGET (priv->addr_list));
+	return FALSE;
+}
+
 static void
 finish_setup (CEPageIP6 *self, gpointer unused, GError *error, gpointer user_data)
 {
@@ -763,6 +908,8 @@ finish_setup (CEPageIP6 *self, gpointer unused, GError *error, gpointer user_dat
 	gtk_tree_view_column_set_expand (GTK_TREE_VIEW_COLUMN (column), TRUE);
 	gtk_tree_view_column_set_clickable (GTK_TREE_VIEW_COLUMN (column), TRUE);
 
+	g_signal_connect (priv->addr_list, "button-press-event", G_CALLBACK (tree_view_button_pressed_cb), self);
+
 	gtk_widget_set_sensitive (GTK_WIDGET (priv->addr_add), TRUE);
 	gtk_widget_set_sensitive (GTK_WIDGET (priv->addr_delete), FALSE);
 
@@ -1001,6 +1148,9 @@ validate (CEPage *page, NMConnection *connection, GError **error)
 static void
 ce_page_ip6_init (CEPageIP6 *self)
 {
+	CEPageIP6Private *priv = CE_PAGE_IP6_GET_PRIVATE (self);
+
+	priv->last_column = -1;
 }
 
 static void
