@@ -17,15 +17,15 @@
  * with this program; if not, write to the Free Software Foundation, Inc.,
  * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  *
- * (C) Copyright 2008 - 2009 Red Hat, Inc.
+ * (C) Copyright 2008 - 2011 Red Hat, Inc.
  */
 
 #include <stdlib.h>
+#include <ctype.h>
 
 #include <glib.h>
 #include <glib/gi18n.h>
 
-#include <glade/glade.h>
 #include <gtk/gtk.h>
 
 #include <NetworkManager.h>
@@ -42,7 +42,7 @@
 #define DEVICE_TAG "device"
 #define TYPE_TAG "setting-type"
 
-static char *get_selected_country (MobileWizard *self);
+static char *get_selected_country (MobileWizard *self, gboolean *unlisted);
 static NmnMobileProvider *get_selected_provider (MobileWizard *self);
 static NmnMobileAccessMethodType get_provider_unlisted_type (MobileWizard *self);
 static NmnMobileAccessMethod *get_selected_method (MobileWizard *self, gboolean *manual);
@@ -272,7 +272,7 @@ confirm_prepare (MobileWizard *self)
 	gboolean manual = FALSE;
 	GString *str;
 
-	country = get_selected_country (self);
+	country = get_selected_country (self, NULL);
 	provider = get_selected_provider (self);
 	if (provider)
 		method = get_selected_method (self, &manual);
@@ -425,6 +425,39 @@ plan_row_separator_func (GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
 	return draw_separator;
 }
 
+static void
+apn_filter_cb (GtkEntry *   entry,
+               const gchar *text,
+               gint         length,
+               gint *       position,
+               gpointer     user_data)
+{
+	GtkEditable *editable = GTK_EDITABLE (entry);
+	int i, count = 0;
+	gchar *result = g_new0 (gchar, length);
+
+	for (i = 0; i < length; i++) {
+		if (   isalnum (text[i])
+		    || (text[i] == '.')
+		    || (text[i] == '_')
+		    || (text[i] == '-'))
+			result[count++] = text[i];
+	}
+
+	if (count > 0) {
+		g_signal_handlers_block_by_func (G_OBJECT (editable),
+		                                 G_CALLBACK (apn_filter_cb),
+		                                 user_data);
+		gtk_editable_insert_text (editable, result, count, position);
+		g_signal_handlers_unblock_by_func (G_OBJECT (editable),
+		                                   G_CALLBACK (apn_filter_cb),
+		                                   user_data);
+	}
+
+	g_signal_stop_emission_by_name (G_OBJECT (editable), "insert-text");
+	g_free (result);
+}
+
 static void
 plan_setup (MobileWizard *self)
 {
@@ -464,7 +497,8 @@ plan_setup (MobileWizard *self)
 
 	self->plan_unlisted_entry = gtk_entry_new ();
 	gtk_label_set_mnemonic_widget (GTK_LABEL (label), self->plan_unlisted_entry);
-	gtk_entry_set_max_length (GTK_ENTRY (self->plan_unlisted_entry), 40);
+	gtk_entry_set_max_length (GTK_ENTRY (self->plan_unlisted_entry), 64);
+	g_signal_connect (self->plan_unlisted_entry, "insert-text", G_CALLBACK (apn_filter_cb), self);
 	g_signal_connect_swapped (self->plan_unlisted_entry, "changed", G_CALLBACK (plan_update_complete), self);
 
 	alignment = gtk_alignment_new (0, 0.5, 0.5, 0);
@@ -747,11 +781,23 @@ providers_setup (MobileWizard *self)
 	gtk_table_attach (GTK_TABLE (unlisted_table), alignment,
 	                  1, 2, 0, 1, GTK_EXPAND | GTK_FILL, 0, 6, 6);
 
+#if GTK_CHECK_VERSION(2,23,0)
+	self->provider_unlisted_type_combo = gtk_combo_box_text_new ();
+#else
 	self->provider_unlisted_type_combo = gtk_combo_box_new_text ();
+#endif
 	gtk_label_set_mnemonic_widget (GTK_LABEL (label), self->provider_unlisted_type_combo);
+#if GTK_CHECK_VERSION(2,23,0)
+	gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (self->provider_unlisted_type_combo),
+#else
 	gtk_combo_box_append_text (GTK_COMBO_BOX (self->provider_unlisted_type_combo),
+#endif
 	                           _("My provider uses GSM technology (GPRS, EDGE, UMTS, HSPA)"));
+#if GTK_CHECK_VERSION(2,23,0)
+	gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (self->provider_unlisted_type_combo),
+#else
 	gtk_combo_box_append_text (GTK_COMBO_BOX (self->provider_unlisted_type_combo),
+#endif
 	                           _("My provider uses CDMA technology (1xRTT, EVDO)"));
 	gtk_combo_box_set_active (GTK_COMBO_BOX (self->provider_unlisted_type_combo), 0);
 
@@ -779,7 +825,14 @@ providers_prepare (MobileWizard *self)
 
 	gtk_tree_store_clear (self->providers_store);
 
-	country = get_selected_country (self);
+	country = get_selected_country (self, NULL);
+	if (!country) {
+		/* Unlisted country */
+		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (self->provider_unlisted_radio), TRUE);
+		gtk_widget_set_sensitive (GTK_WIDGET (self->providers_view_radio), FALSE);
+		goto done;
+	}
+
 	providers = g_hash_table_lookup (self->providers, country);
 	g_free (country);
 
@@ -837,6 +890,7 @@ providers_prepare (MobileWizard *self)
 		}
 	}
 
+done:
 	providers_radio_toggled (NULL, self);
 
 	/* Initial completeness state */
@@ -914,12 +968,13 @@ add_one_country (gpointer key, gpointer value, gpointer user_data)
 }
 
 static char *
-get_selected_country (MobileWizard *self)
+get_selected_country (MobileWizard *self, gboolean *out_unlisted)
 {
 	GtkTreeSelection *selection;
 	GtkTreeModel *model = NULL;
 	GtkTreeIter iter;
 	char *country = NULL;
+	gboolean unlisted = FALSE;
 
 	if (!self->country_codes)
 		return NULL;
@@ -927,8 +982,15 @@ get_selected_country (MobileWizard *self)
 	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (self->country_view));
 	g_assert (selection);
 
-	if (gtk_tree_selection_get_selected (GTK_TREE_SELECTION (selection), &model, &iter))
-		gtk_tree_model_get (model, &iter, 0, &country, -1);
+	if (gtk_tree_selection_get_selected (GTK_TREE_SELECTION (selection), &model, &iter)) {
+		gtk_tree_model_get (model, &iter, 0, &country, 1, &unlisted, -1);
+		if (unlisted) {
+			g_free (country);
+			country = NULL;
+			if (out_unlisted)
+				*out_unlisted = unlisted;
+		}
+	}
 
 	return country;
 }
@@ -937,14 +999,51 @@ static void
 country_update_complete (MobileWizard *self)
 {
 	char *country = NULL;
+	gboolean unlisted = FALSE;
 
-	country = get_selected_country (self);
+	country = get_selected_country (self, &unlisted);
 	gtk_assistant_set_page_complete (GTK_ASSISTANT (self->assistant),
 	                                 self->country_page,
-	                                 !!country);
+	                                 (!!country || unlisted));
 	g_free (country);
 }
 
+static gint
+country_sort_func (GtkTreeModel *model,
+                   GtkTreeIter *a,
+                   GtkTreeIter *b,
+                   gpointer user_data)
+{
+	char *a_str = NULL, *b_str = NULL;
+	gboolean a_def = FALSE, b_def = FALSE;
+	gint ret = 0;
+
+	gtk_tree_model_get (model, a, 0, &a_str, 1, &a_def, -1);
+	gtk_tree_model_get (model, b, 0, &b_str, 1, &b_def, -1);
+
+	if (a_def) {
+		ret = -1;
+		goto out;
+	} else if (b_def) {
+		ret = 1;
+		goto out;
+	}
+
+	if (a_str && !b_str)
+		ret = -1;
+	else if (!a_str && b_str)
+		ret = 1;
+	else if (!a_str && !b_str)
+		ret = 0;
+	else
+		ret = g_utf8_collate (a_str, b_str);
+
+out:
+	g_free (a_str);
+	g_free (b_str);
+	return ret;
+}
+
 static void
 country_setup (MobileWizard *self)
 {
@@ -952,14 +1051,15 @@ country_setup (MobileWizard *self)
 	GtkCellRenderer *renderer;
 	GtkTreeViewColumn *column;
 	GtkTreeSelection *selection;
+	GtkTreeIter unlisted_iter;
 
 	vbox = gtk_vbox_new (FALSE, 6);
 	gtk_container_set_border_width (GTK_CONTAINER (vbox), 12);
-	label = gtk_label_new (_("Country List:"));
+	label = gtk_label_new (_("Country or Region List:"));
 	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
 	gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, TRUE, 0);
 
-	self->country_store = gtk_tree_store_new (1, G_TYPE_STRING);
+	self->country_store = gtk_tree_store_new (2, G_TYPE_STRING, G_TYPE_BOOLEAN);
 
 	self->country_sort = GTK_TREE_MODEL_SORT (gtk_tree_model_sort_new_with_model (GTK_TREE_MODEL (self->country_store)));
 	gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE (self->country_sort), 0, GTK_SORT_ASCENDING);
@@ -967,10 +1067,23 @@ country_setup (MobileWizard *self)
 	self->country_view = gtk_tree_view_new_with_model (GTK_TREE_MODEL (self->country_sort));
 
 	renderer = gtk_cell_renderer_text_new ();
-	column = gtk_tree_view_column_new_with_attributes (_("Country"), renderer, "text", 0, NULL);
+	column = gtk_tree_view_column_new_with_attributes (_("Country or region"), renderer, "text", 0, NULL);
 	gtk_tree_view_append_column (GTK_TREE_VIEW (self->country_view), column);
 	gtk_tree_view_column_set_clickable (column, TRUE);
 
+	/* My country is not listed... */
+	gtk_tree_store_append (GTK_TREE_STORE (self->country_store), &unlisted_iter, NULL);
+	gtk_tree_store_set (GTK_TREE_STORE (self->country_store), &unlisted_iter,
+	                    0, _("My country is not listed"),
+	                    1, TRUE,
+	                    -1);
+	gtk_tree_sortable_set_sort_func (GTK_TREE_SORTABLE (self->country_sort),
+	                                 0,
+	                                 country_sort_func,
+	                                 NULL,
+	                                 NULL);
+
+	/* Add the rest of the providers */
 	if (self->providers)
 		g_hash_table_foreach (self->providers, add_one_country, self);
 	g_object_set (G_OBJECT (self->country_view), "enable-search", TRUE, NULL);
@@ -1007,7 +1120,7 @@ country_setup (MobileWizard *self)
 	gtk_box_pack_start (GTK_BOX (vbox), alignment, TRUE, TRUE, 6);
 
 	self->country_idx = gtk_assistant_append_page (GTK_ASSISTANT (self->assistant), vbox);
-	gtk_assistant_set_page_title (GTK_ASSISTANT (self->assistant), vbox, _("Choose your Provider's Country"));
+	gtk_assistant_set_page_title (GTK_ASSISTANT (self->assistant), vbox, _("Choose your Provider's Country or Region"));
 	gtk_assistant_set_page_type (GTK_ASSISTANT (self->assistant), vbox, GTK_ASSISTANT_PAGE_CONTENT);
 	gtk_assistant_set_page_complete (GTK_ASSISTANT (self->assistant), vbox, TRUE);
 	gtk_widget_show_all (vbox);
